/* 32-bit integer traits */

#define Scalar int32       /* integer type */
#define Int int32          /* corresponding signed integer type */
#define UInt uint32        /* corresponding unsigned integer type */
#define PBITS 5            /* number of bits needed to encode precision */
#define NBMASK 0xaaaaaaaau /* negabinary mask */
